#
# Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#

"""
Module: config_processor

This module provides the function generate_device_cfg(), which processes a configuration
file (YAML or legacy .cfg) and generates per-device INI-style configuration files 
(saving them to /tmp). The generated file is then assigned to the device's 
'device_config_file' attribute.

Usage in your main installer (doca-installer.py):

    if args.config_file:
        generate_device_cfg(bf_devices_upgrade_candidate, args.config_file)
"""

import os
import yaml
import re
import sys

def _legacy_config_processor(config_file_path: str, rshim: str) -> str:
    """
    Process a legacy configuration file and generate an INI-style configuration file
    specific to a given rshim device.

    Args:
        config_file_path (str): The path to the legacy configuration file.
        rshim (str): The rshim device to generate the configuration file for.

    Returns:
        str: The content of the generated configuration file.
    """
    try:
        with open(config_file_path, "r") as f:
            full_cfg = f.read()
    except FileNotFoundError:
        print(f"ERROR: File '{config_file_path}' not found.")
        sys.exit(1)
    rshim_config_vars = {}
    rshim_config_methods = {}
    rshim_number = rshim.upper()
    adjusted_lines = []
    rshim_config_vars[rshim_number] = {}
    rshim_config_methods[rshim_number] = {}
    full_cfg_lines = full_cfg.splitlines()

    for line in full_cfg_lines:
        if line.startswith(f"{rshim_number}_") and "=" in line:
            # Remove the prefix and keep the variable
            match = re.match(rf"({rshim_number}_)?(\w+)=(.*)", line)
            if match:
                prefix, var_name, value = match.groups()
                # Add to the RSHIM-specific dictionary if not already present
                if var_name not in rshim_config_vars[rshim_number]:
                    rshim_config_vars[rshim_number][
                        var_name
                    ] = value.strip()

        elif line.startswith(f"{rshim_number}_") and "()" in line:
            # Capture the method name without the prefix
            match = re.match(rf"{rshim_number}_(\w+)\(\)", line)
            if match:
                method_name = match.group(1)
                # Generate a new method that calls the original method
                if method_name not in rshim_config_methods[rshim_number]:
                    rshim_config_methods[rshim_number][
                        method_name
                    ] = f"{method_name}()\n{{\n    {line.strip().replace('()', '')}\n}}"
    adjusted_lines = [
        f"{key}={value}"
        for key, value in rshim_config_vars[rshim_number].items()
    ]
    adjusted_lines += rshim_config_methods[rshim_number].values()
    remaining_lines = [
        line
        for line in full_cfg_lines
        if not any(
            re.match(rf"({rshim_number}_)?{key}(=|\\(\\))", line)
            for key in rshim_config_vars[rshim_number]
        )
    ]
    device_cfg_content = "\n".join(
        adjusted_lines
        + ["", "# ORIGINAL-CONFIG-CONTENT", ""]
        + remaining_lines
    )
    return device_cfg_content

def _load_yaml_config(config_file_path: str) -> dict:
    """Load and parse a YAML configuration file."""
    with open(config_file_path, 'r') as f:
        return yaml.safe_load(f)

def _generate_ini_content_from_yaml(config_yaml: dict, rshim_id: str) -> str:
    """
    Generate an INI-style configuration from the YAML configuration for a specific rshim device.
    
    The YAML configuration is expected to have:
      - A top-level "global_configs" section (which may contain "variables" and "methods")
      - A top-level "rshim" section with device-specific configuration.
    
    Global and device-specific settings are merged; if the same key is defined in both,
    the device-specific definition takes priority.
    """
    # Global configuration: variables and optionally methods.
    global_vars = {}
    global_methods = {}
    if 'global_configs' in config_yaml:
        gc = config_yaml['global_configs']
        if isinstance(gc, dict):
            global_vars = gc.get('variables', {})
            global_methods = gc.get('methods', {})

    # Device-specific configuration.
    device_config = config_yaml.get('rshim', {}).get(rshim_id, {})
    device_vars = device_config.get('variables', {})
    device_methods = device_config.get('methods', {})

    # Merge definitions (device-specific takes priority).
    # Thanks to unpacking device_vars/methods second, they will override global_vars/methods.
    merged_vars = {**global_vars, **device_vars}
    merged_methods = {**global_methods, **device_methods}

    lines = []

    # Create variable assignments.
    for key, value in merged_vars.items():
        if isinstance(value, str) and not (value.startswith('"') and value.endswith('"')):
            value_str = f"\"{value}\""
        else:
            value_str = str(value)
        lines.append(f"{key}={value_str}")

    lines.append("")  # Blank line to separate variables from methods.

    # Create shell function definitions.
    for method_name, method_details in merged_methods.items():
        lines.append(f"{method_name}()")
        lines.append("{")
        command = method_details.get("command", "")
        # Split and preserve the newlines from the YAML
        command_lines = command.splitlines()
        for cmd_line in command_lines:
            # Add proper indentation for each line
            lines.append(f"    {cmd_line.rstrip()}")
        lines.append("}")
    
    return "\n".join(lines)

def generate_device_cfg(bf_devices: list, config_file_path: str) -> None:
    """
    Generate per-device INI-style configuration files based on the given configuration file.

    For a YAML file:
      - Loads the YAML configuration.
      - For each device (which must have a 'rshim_device' attribute),
        merges the global settings with device-specific settings.
      - Writes the resulting INI-style content to
        /tmp/{<device.bf_pci_address with colons replaced>}_device_config.cfg.
      - Updates the device's 'device_config_file' attribute.

    For a legacy .cfg file:
      - Reads the file content and writes a copy for each device.
    """
    _, ext = os.path.splitext(config_file_path)
    ext = ext.lower()

    if ext in ['.yaml', '.yml']:
        config_yaml = _load_yaml_config(config_file_path)
        for bf_device in bf_devices:
            rshim_id = getattr(bf_device, "rshim_device", None)
            if not rshim_id:
                print(f"ERROR: Device {bf_device.bf_pci_address} is missing a 'rshim_device' attribute.")
                continue
            content = _generate_ini_content_from_yaml(config_yaml, rshim_id)
            output_file = f"/tmp/{bf_device.rshim_device}-bf.cfg"
            with open(output_file, "w") as f:
                f.write(content)
            bf_device.device_config_file = output_file
            print(f"Generated configuration for {bf_device.bf_pci_address} at {output_file}")

    elif ext == '.cfg':
        # Legacy handling: simply copy the contents for each bf_device.
        for bf_device in bf_devices:
            content = _legacy_config_processor(config_file_path, bf_device.rshim_device)
            output_file = f"/tmp/{bf_device.rshim_device}-bf.cfg"
            with open(output_file, "w") as f:
                f.write(content)
            bf_device.device_config_file = output_file
            print(f"Generated configuration for {bf_device.bf_pci_address} at {output_file}")
    else:
        raise ValueError(f"Unsupported configuration file format: {ext}")